# Copyright 2021 Max Beinlich

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum

class MarketState(Enum):
    """
    The MarketState shows what state the instruments market currently assumes.

    Notice that this does not include a "CLOSED" market state. This is because 
    quotes are not emitted when the underlying market is closed / when no 
    trading occures.

    Depending on the MarketState quotes might be less frequent / delayed.
    """

    PRE         = 0
    REGULAR     = 1
    POST        = 2
    EXTENDED    = 3
    NONE        = None


# Copyright 2021 Max Beinlich

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

import uuid

from yflive.enums import MarketState, OptionType, QuoteType

class Quote:
    """
    The Quote object is provided for handling market data in a structured way.

    Quote instances are emitted in real-time as they are received from 
    Yahoo! Finance and offer a common interface for acquiring the data 
    provided.
    """

    # DO NOT CHANGE ORDER
    __fields__ = ["identifier", "price", "time", "currency", "exchange", 
                  "quoteType", "marketState", "changePercent", "dayVolume", 
                  "dayHigh", "dayLow", "change", "shortName", 
                  "expireDate", "openPrice", "previousClose", "strikePrice", 
                  "underlyingSymbol", "openInterest", "optionType", 
                  "miniOption", "lastSize", "bid", "bidSize", "ask", "askSize", 
                  "priceHint", "vol_24hr", "volAllCurrencies", "fromCurrency", 
                  "lastMarket", "circulatingSupply", "marketCap"]

    def __init__(self, **kwargs):
        """
        Initialize new Quote object.

        Parameters:
        -----------
        identifier: <type>
            Unique Yahoo! Finance identifier
        time: <type>
            Time of quote
        quoteType: <type>
            Type of underlying instrument
        kwargs:
            Additional quote information (limited by __fields__)
        """
        self._uuid = str(uuid.uuid4())

        for f in self.__fields__:
            setattr(self, f, kwargs.get(f))

        # Mandatory
        self.identifier = str(kwargs["identifier"]).upper()
        self.time = kwargs["time"]
        self.quoteType = QuoteType(kwargs["quoteType"])

        # Enum redeclaration
        self.marketState = MarketState(self.marketState)
        self.optionType = OptionType(self.optionType)

    def __str__(self): 
        return "{0} {1} - Price: {2}, {3} : {4}".format(
                    self.quoteType.name, self.identifier, 
                    self.price, self.exchange,
                    self.marketState.name
                )
        


# Copyright 2021 Max Beinlich

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum

class OptionType(Enum):
    """
    The type of option, if quoteType is OPTION.

    If a quote assumes the type OPTION, the OptionType should either be CALL or
    PUT. The OptionType defaults to OptionType.NONE.
    """

    CALL    = 0
    PUT     = 1
    NONE    = None


# Copyright 2021 Max Beinlich

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum

class QuoteType(Enum):
    """
    The QuoteType indicates the type of financial instrument represented by a
    quote.

    Depending on the QuoteType, a Quote might contain different data,
    thus the QuoteType is used for communicating what should be expected in 
    a Quote. 
    
    Something to watch out for, is that yahoo finance only offers delayed 
    quotes for certain QuoteTypes.
    """

    UNDEFINED       = 0
    ALTSYMBOL       = 5
    HEARTBEAT       = 7
    EQUITY          = 8         
    INDEX           = 9         
    MUTUALFUND      = 11
    MONEYMARKET     = 12
    OPTION          = 13        
    CURRENCY        = 14        
    WARRENT         = 15        
    BOND            = 17        
    FUTURE          = 18
    ETF             = 20        
    COMMODITY       = 23
    ECNQUOTE        = 28
    CRYPTOCURRENCY  = 41
    INDICATOR       = 42
    INDUSTRY        = 1000
    NONE            = None


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: yfquote.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='yfquote.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\ryfquote.proto\"\x84\x05\n\x07YFQuote\x12\x12\n\nidentifier\x18\x01 \x01(\t\x12\r\n\x05price\x18\x02 \x01(\x02\x12\x0c\n\x04time\x18\x03 \x01(\x12\x12\x10\n\x08\x63urrency\x18\x04 \x01(\t\x12\x10\n\x08\x65xchange\x18\x05 \x01(\t\x12\x11\n\tquoteType\x18\x06 \x01(\x05\x12\x13\n\x0bmarketState\x18\x07 \x01(\x05\x12\x15\n\rchangePercent\x18\x08 \x01(\x02\x12\x0e\n\x06\x63hange\x18\x0c \x01(\x02\x12\x11\n\tdayVolume\x18\t \x01(\x12\x12\x0f\n\x07\x64\x61yHigh\x18\n \x01(\x02\x12\x0e\n\x06\x64\x61yLow\x18\x0b \x01(\x02\x12\x11\n\topenPrice\x18\x0f \x01(\x02\x12\x15\n\rpreviousClose\x18\x10 \x01(\x02\x12\x11\n\tshortName\x18\r \x01(\t\x12\x12\n\nexpireDate\x18\x0e \x01(\x12\x12\x13\n\x0bstrikePrice\x18\x11 \x01(\x02\x12\x18\n\x10underlyingSymbol\x18\x12 \x01(\t\x12\x14\n\x0copenInterest\x18\x13 \x01(\x12\x12\x12\n\noptionType\x18\x14 \x01(\x12\x12\x12\n\nminiOption\x18\x15 \x01(\x12\x12\x10\n\x08lastSize\x18\x16 \x01(\x12\x12\x0b\n\x03\x62id\x18\x17 \x01(\x02\x12\x0f\n\x07\x62idSize\x18\x18 \x01(\x12\x12\x0b\n\x03\x61sk\x18\x19 \x01(\x02\x12\x0f\n\x07\x61skSize\x18\x1a \x01(\x12\x12\x11\n\tpriceHint\x18\x1b \x01(\x12\x12\x10\n\x08vol_24hr\x18\x1c \x01(\x12\x12\x18\n\x10volAllCurrencies\x18\x1d \x01(\x12\x12\x14\n\x0c\x66romCurrency\x18\x1e \x01(\t\x12\x12\n\nlastMarket\x18\x1f \x01(\t\x12\x19\n\x11\x63irculatingSupply\x18  \x01(\x01\x12\x11\n\tmarketCap\x18! \x01(\x01\x62\x06proto3'
)




_YFQUOTE = _descriptor.Descriptor(
  name='YFQuote',
  full_name='YFQuote',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='identifier', full_name='YFQuote.identifier', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='price', full_name='YFQuote.price', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='time', full_name='YFQuote.time', index=2,
      number=3, type=18, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='currency', full_name='YFQuote.currency', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='exchange', full_name='YFQuote.exchange', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='quoteType', full_name='YFQuote.quoteType', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='marketState', full_name='YFQuote.marketState', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='changePercent', full_name='YFQuote.changePercent', index=7,
      number=8, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='change', full_name='YFQuote.change', index=8,
      number=12, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dayVolume', full_name='YFQuote.dayVolume', index=9,
      number=9, type=18, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dayHigh', full_name='YFQuote.dayHigh', index=10,
      number=10, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dayLow', full_name='YFQuote.dayLow', index=11,
      number=11, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='openPrice', full_name='YFQuote.openPrice', index=12,
      number=15, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='previousClose', full_name='YFQuote.previousClose', index=13,
      number=16, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='shortName', full_name='YFQuote.shortName', index=14,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='expireDate', full_name='YFQuote.expireDate', index=15,
      number=14, type=18, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='strikePrice', full_name='YFQuote.strikePrice', index=16,
      number=17, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='underlyingSymbol', full_name='YFQuote.underlyingSymbol', index=17,
      number=18, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='openInterest', full_name='YFQuote.openInterest', index=18,
      number=19, type=18, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='optionType', full_name='YFQuote.optionType', index=19,
      number=20, type=18, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='miniOption', full_name='YFQuote.miniOption', index=20,
      number=21, type=18, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='lastSize', full_name='YFQuote.lastSize', index=21,
      number=22, type=18, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bid', full_name='YFQuote.bid', index=22,
      number=23, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bidSize', full_name='YFQuote.bidSize', index=23,
      number=24, type=18, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ask', full_name='YFQuote.ask', index=24,
      number=25, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='askSize', full_name='YFQuote.askSize', index=25,
      number=26, type=18, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='priceHint', full_name='YFQuote.priceHint', index=26,
      number=27, type=18, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='vol_24hr', full_name='YFQuote.vol_24hr', index=27,
      number=28, type=18, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='volAllCurrencies', full_name='YFQuote.volAllCurrencies', index=28,
      number=29, type=18, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fromCurrency', full_name='YFQuote.fromCurrency', index=29,
      number=30, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='lastMarket', full_name='YFQuote.lastMarket', index=30,
      number=31, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='circulatingSupply', full_name='YFQuote.circulatingSupply', index=31,
      number=32, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='marketCap', full_name='YFQuote.marketCap', index=32,
      number=33, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18,
  serialized_end=662,
)

DESCRIPTOR.message_types_by_name['YFQuote'] = _YFQUOTE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

YFQuote = _reflection.GeneratedProtocolMessageType('YFQuote', (_message.Message,), {
  'DESCRIPTOR' : _YFQUOTE,
  '__module__' : 'yfquote_pb2'
  # @@protoc_insertion_point(class_scope:YFQuote)
  })
_sym_db.RegisterMessage(YFQuote)


# @@protoc_insertion_point(module_scope)



# Copyright 2021 Max Beinlich

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yflive.quote import Quote
from yflive.streamer import QuoteStreamer

from yflive.enums.market_state import MarketState
from yflive.enums.option_type import OptionType
from yflive.enums.quote_type import QuoteType

__version__ = "0.1.5"
__author__ = "Max Beinlich"

__all__ = ['QuoteType', 'MarketState', 'OptionType', 'Quote', 'QuoteStreamer']


# Copyright 2021 Max Beinlich

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

import base64

from yflive.yfquote_pb2 import YFQuote
from yflive.quote import Quote

class QuoteReader:
    """
    Reader class for Yahoo! Finance websocket messages.

    This class implements the logic needed to decrypt websocket messages sent
    by Yahoo! Finance.
    """

    def __init__(self, buf: List[int], pos: int, length: int):
        self.buf = buf
        self.pos = pos
        self.length = length

    def incr(self):
        self.pos += 1
        return self.pos - 1

    def uint32(self):
        value = 4294967295
        value = (self.buf[self.pos] & 127) >> 0
        if (self.buf[self.incr()] < 128): 
            return value

        value = (value | (self.buf[self.pos] & 127) << 7) >> 0
        if (self.buf[self.incr()] < 128): 
            return value

        value = (value | (self.buf[self.pos] & 127) << 14) >> 0
        if (self.buf[self.incr()] < 128): 
            return value

        value = (value | (self.buf[self.pos] & 127) << 21) >> 0 
        if (self.buf[self.incr()] < 128): 
            return value

        value = (value | (self.buf[self.pos] &  15) << 28) >> 0 
        if (self.buf[self.incr()] < 128): 
            return value

        self.pos += 5
        if ((self.pos) > self.length):
            self.pos = self.length
        return value

    def skip(self, length=None):
        if isinstance(length, int):
            if self.pos + length > self.length:
                return IndexError("index out of range")
            self.pos += length
        else:
            while True:
                if self.pos >= self.length:
                    return IndexError("index out of range")
                if self.buf[self.incr()] & 128 == 0:
                    break
        return

    def skipType(self, wireType):
        if wireType == 0:
            self.skip()
        elif wireType == 1:
            self.skip(8)
        elif wireType == 2:
            self.skip(self.uint32())
        elif wireType == 3:
            wireType = self.uint32() & 7
            while wireType != 4:
                self.skipType(wireType)
                wireType = self.uint32() & 7
        elif wireType == 5:
            self.skip(4)
        else:
            return 
        return

    # ==========================================================================
    # Parse Yahoo! Finance websocket message to Quote
    # ==========================================================================

    @staticmethod
    def parse(msg: str):
        """
        Parse Yahoo! Finance message to Quote object
        
        Parameters:
        -----------
        msg: str
            Yahoo! Finance base64 encoded quote string
        """
        message_bytes = base64.b64decode(msg)
        yfquote = YFQuote()
        yfquote.ParseFromString(message_bytes)

        fields = {}
        for f in QuoteReader.available_fields(msg):
            fields[f] = getattr(yfquote, f, None)

        return Quote(**fields)

    @staticmethod
    def available_fields(msg: str): 
        """
        Get available fields from message

        Parameters:
        -----------
        msg: str
            Yahoo! Finance base64 encoded quote string
        """
        buffer = list(base64.b64decode(msg))
        reader = QuoteReader(buffer, 0, len(buffer))
        c = reader.length
        available_fields = []
        while reader.pos < c:
            t = reader.uint32()
            available_fields.append(Quote.__fields__[(t >> 3) - 1])
            reader.skipType(t & 7)
        return available_fields


# Copyright 2021 Max Beinlich

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

import json
import ssl
import threading
import logging

import websocket as ws

from ._reader import QuoteReader

__all__ = ['YAHOO_FINANCE_SOCKET', 'QuoteStreamer']

_logger = logging.getLogger("yflive")

# ==============================================================================
# QuoteStreamer
# ==============================================================================

YAHOO_FINANCE_SOCKET = "wss://streamer.finance.yahoo.com/"

class QuoteStreamer:
    """
    The QuoteStreamer streams live quote data from yahoo!finance.

    In order to receive live data we connect to the yahoo!finance websocket and
    subscribe to certain identifiers.

    The websocket responds with quotes of the financial instruments previously 
    subscribed to, which are then parced.
    """

    on_connect = None
    on_quote = None
    on_error = None
    on_close = None

    instance = None

    def __init__(self, enable_trace: bool=False):
        """Get QuoteStreamer singleton"""
        self._subscribed = set()
        self._websocket = None

        ws.enableTrace(enable_trace)

        self._ws_thread = None

    def __new__(cls, **kwargs):
        if not cls.instance:
            cls.instance = super(QuoteStreamer, cls).__new__(cls, **kwargs)
        return cls.instance

    def __del__(self):
        self.stop()

    def start(self, blocking=True):
        """
        Connect to the yahoo!finance websocket.

        Establish a connection to the yahoo!finance websocket with given
        callback methods.

        Parameters:
        -----------
        blocking: bool 
            True <default>
            False -> Run on non blocking Thread
        """
        if blocking is False:
            self._ws_thread = threading.Thread(target=self._run)
            self._ws_thread.daemon = True
            self._ws_thread.start()
        else: 
            self._run()


    def stop(self):
        """
        Disconnect the Yahoo! Finance websocket.
        """
        if not self.streaming: 
            return
        _logger.debug("Stopping QuoteStreamer...")
        if isinstance(self._websocket, ws.WebSocketApp):
            self._websocket.close()
            self._websocket = None
        if self._ws_thread:
            self._ws_thread.join()

    def _run(self):
        try:
            self._websocket = ws.WebSocketApp(
                        YAHOO_FINANCE_SOCKET, 
                        on_error = _ws_error, 
                        on_close = _ws_close, 
                        on_message = _ws_message,
                        on_open = _ws_open)
            self._websocket.run_forever(sslopt={"cert_reqs": ssl.CERT_NONE})
        except (Exception, KeyboardInterrupt, SystemExit) as e:
            self._callback(self.on_error, (e))
            if isinstance(e, SystemExit):
                # propagate SystemExit further
                raise
            self.stop()
            return not isinstance(e, KeyboardInterrupt)

    @property
    def subscribed(self) -> List:
        """Get all currently tracked identifiers."""
        return list(self._subscribed)

    @property
    def streaming(self) -> bool:
        """Get current streaming state."""
        return self._websocket is not None

    def subscribe(self, identifiers=None):
        """
        Subscribe to identifiers.
        
        Parameters:
        -----------
        identifiers: <type>
            identifiers to subscribe to
        """
        if identifiers is None and len(identifiers) <= 0:
            return
        identifiers = set(identifiers) - self._subscribed
        self._subscribed = self._subscribed | identifiers
        _logger.debug(f"Subscribing to {list(identifiers)}")
        if self.streaming and len(identifiers) > 0:
            msg = json.dumps({"subscribe": list(identifiers)})
            self._websocket.send(msg)
        return

    def unsubscribe(self, identifiers=None):
        """
        Unsubscribe from identifiers.
        
        Parameters:
        -----------
        identifiers: <type>
            identifiers to unsubscribe from
        """
        if identifiers is None and len(identifiers) <= 0:
            return
        identifiers = self._subscribed & set(identifiers)
        self._subscribed = self._subscribed - identifiers
        _logger.debug(f"Unsubscribing from {list(identifiers)}")
        if self.streaming and len(identifiers) > 0:
            msg = json.dumps({"unsubscribe": list(identifiers)})
            self._websocket.send(msg)
        return

    def _callback(self, callback, *args):
        if callback and callable(callback):
            try:
                callback(self, *args)
            except Exception as e:
                _logger.error("error from callback {}: {}".format(callback, e))

# ==============================================================================
# Websocket callback methods
# ==============================================================================

_streamer = QuoteStreamer()

def _ws_open(ws):
    _logger.debug("Yahoo! Finance connection opened")
    _streamer._callback(_streamer.on_connect)
    if len(_streamer.subscribed) > 0:
        msg = json.dumps({"subscribe": list(_streamer.subscribed)})
        ws.send(msg)

def _ws_message(ws, message):
    quote = QuoteReader.parse(message)
    _logger.info(f"Quote received: {str(quote)}")
    _streamer._callback(_streamer.on_quote, (quote))
    
def _ws_error(ws, error):
    _logger.error(f"Error encountered: {error}")
    _streamer._callback(_streamer.on_error, (error))

def _ws_close(ws):
    _logger.debug("Connection closed")
    _streamer._callback(_streamer.on_close)
    _streamer.stop()


import setuptools

import io
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with io.open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    name="yflive",
    version="0.1.5",
    description="live Yahoo! Finance data streamer",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Max Beinlich",
    author_email="maxbeinlich@gmail.com",
    license="Apache",
    url='https://github.com/mbnlc/yflive.git',
    download_url='https://github.com/mbnlc/yflive/releases',
    project_urls={
        "Source": "https://github.com/mbnlc/yflive",
        "Bug Tracker": "https://github.com/mbnlc/yflive/issues",
    },
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 5 - Production/Stable",

        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",

        "Operating System :: OS Independent",

        "License :: OSI Approved :: Apache Software License",

        "Intended Audience :: Developers",
        "Topic :: Office/Business :: Financial",
        "Topic :: Office/Business :: Financial :: Investment",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    install_requires=["websocket-client", "protobuf"],
    packages=setuptools.find_packages(exclude="tests"),
    include_package_data=True,
    python_requires=">=3.6"
)